﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace DigitoalNL.VELReader
{
    public class VELReader
    {
        enum ElFileTypes
        {
            INVALID = 0,
            INFO = 1,
            DATA = 2
        }

        public VELReader()
        {
            Info = new InfoClass();
            Log = new List<LogLine>();
        }

        public InfoClass Info;
        public class InfoClass
        {
            private byte[] _bBuffer = new byte[102];

            public InfoClass() { }
            public InfoClass(string fileName)
            {
                Read(fileName);
            }

            public void Read(string fileName)
            {
                BinaryReader binRdr = new BinaryReader(File.Open(fileName, FileMode.Open));
                byte[] bInfo = new byte[102];
                int i = binRdr.Read(bInfo, 0, 102);

                _bBuffer = new byte[bInfo.Length];
                _bBuffer = bInfo;

                binRdr.Close();
            }

            /// <summary>
            /// Get: return total power consumed in kiloWatthour
            /// </summary>
            public decimal TotalPowerConsumed
            {
                get
                {
                    byte[] bBuff = new byte[]{
                        0x00,
                        _bBuffer[5],
                        _bBuffer[6],
                        _bBuffer[7]};
                    
                    if (BitConverter.IsLittleEndian)
                        Array.Reverse(bBuff);
                    decimal iWattHour = (BitConverter.ToInt32(bBuff, 0));

                    return iWattHour / 1000;//Return kWh
                }
            }

            /// <summary>
            /// Get: returns hours, 0.5 = 30 minutes, 0.75 = 45 minutes
            /// </summary>
            public decimal TotalRecoredTime
            {
                get
                {
                    byte[] bBuff = new byte[] { 
                        0x00, 
                        _bBuffer[8], 
                        _bBuffer[9], 
                        _bBuffer[10] };
                    if (BitConverter.IsLittleEndian)
                        Array.Reverse(bBuff);

                    int iValue = BitConverter.ToInt32(bBuff, 0);

                    return (decimal)((decimal)iValue / 100);
                }
            }
            
            /// <summary>
            /// Get: returns hours, 0.5 = 30 minutes, 0.75 = 45 minutes
            /// </summary>
            public decimal TotalOnTime
            {
                get
                {
                    byte[] bBuff = new byte[]{
                        0x00,
                        _bBuffer[11],
                        _bBuffer[12],
                        _bBuffer[13]};
                    if (BitConverter.IsLittleEndian)
                        Array.Reverse(bBuff);

                    int iValue = BitConverter.ToInt32(bBuff, 0);

                    return (decimal)((decimal)iValue / 100);
                }
            }
            
            /// <summary>
            /// Get: returns Unit ID (0-9)
            /// </summary>
            public int UnitID
            {
                get
                {
                    byte[] bBuff = new byte[]{
                        0x00,
                        0x00,
                        0x00,
                        _bBuffer[84]};
                    if (BitConverter.IsLittleEndian)
                        Array.Reverse(bBuff);
                    return BitConverter.ToInt32(bBuff, 0);
                }
            }
            
            /// <summary>
            /// Get: Tariff1
            /// </summary>
            public decimal Tariff1
            {
                get
                {
                    string sVal0 = _bBuffer[85].ToString();
                    string sVal1 = _bBuffer[86].ToString();
                    string sVal2 = _bBuffer[87].ToString();
                    string sVal3 = _bBuffer[88].ToString();

                    string sJoin = sVal0 + sVal1 + sVal2 + sVal3;
                    int iValue = Convert.ToInt32(sJoin);
                    return (decimal)((decimal)iValue / 1000);
                }
            }
            
            /// <summary>
            /// Get: Tariff2
            /// </summary>
            public decimal Tariff2
            {
                get
                {
                    string sVal0 = _bBuffer[89].ToString();
                    string sVal1 = _bBuffer[90].ToString();
                    string sVal2 = _bBuffer[91].ToString();
                    string sVal3 = _bBuffer[92].ToString();

                    string sJoin = sVal0 + sVal1 + sVal2 + sVal3;
                    int iValue = Convert.ToInt32(sJoin);
                    return (decimal)((decimal)iValue / 1000);
                }
            }

            public decimal TotalPowerOnTariff1
            {
                get
                {
                    return TotalPowerConsumed * Tariff1;
                }
            }

            public decimal TotalPowerOnTariff2
            {
                get
                {
                    return TotalPowerConsumed * Tariff2;
                }
            }
        }


        
        public void Read(string file)
        {
            Console.WriteLine("ElReader: Read({0})", file);
            ReadFile(file);
        }



        private void ReadFile(string sFilename)
        {
            if (!File.Exists(sFilename)) return;

            BinaryReader binReader = new BinaryReader(File.Open(sFilename, FileMode.Open));

            ElFileTypes eFileType = 0;

            try
            {
                //Match INFO file
                //0x49, 0x4E, 0x46, 0x4F, 0x3A
                //Match DATA file
                //0xE0, 0xC5, 0xEA

                byte[] bHeader = new byte[5];
                int iCnt = binReader.Read(bHeader, 0, 5);
                
                //Print first 5 bytes
                string HexString = string.Empty;
                for (int i = 0; i < iCnt; i++)
                {
                    HexString += string.Format("{0:X2} ", bHeader[i]);
                }
                System.Diagnostics.Debug.WriteLine("File header: " + HexString);
 

                if (iCnt == 5)
                {
                    if (
                        bHeader[0] == 0x49 &&
                        bHeader[1] == 0x4E &&
                        bHeader[2] == 0x46 &&
                        bHeader[3] == 0x4F &&
                        bHeader[4] == 0x3A)
                    {
                        System.Diagnostics.Debug.WriteLine("ElReader: Filetype: INFO");
                        eFileType = ElFileTypes.INFO;
                    }
                    else if (
                        bHeader[0] == 0xE0 &&
                        bHeader[1] == 0xC5 &&
                        bHeader[2] == 0xEA)
                    {
                        System.Diagnostics.Debug.WriteLine("ElReader: Filetype: DATA");
                        eFileType = ElFileTypes.DATA;
                    }
                    else
                    {
                        System.Diagnostics.Debug.WriteLine("ElReader: Invalid file header");
                    }
                }
                else
                    System.Diagnostics.Debug.WriteLine("ElReader: File to short");                
            }
            catch (EndOfStreamException e)
            {
                System.Diagnostics.Debug.WriteLine("EndOfStreamException " + e.Message);
            }
            finally
            {
                binReader.Close();
            }

            switch ((ElFileTypes) eFileType)
            {
                case ElFileTypes.INVALID:
                    return;
                case ElFileTypes.INFO:
                    Info.Read(sFilename);
                    break;
                case ElFileTypes.DATA:
                    ReadDataFile(sFilename);
                    break;
                default:
                    break;
            }            
        }

        public class LogLine
        {
            private DateTime _dtLogTime = DateTime.Now;
            private DateTime _dtFileTime = DateTime.MinValue;
            private decimal _dVoltage = 0;
            private decimal _dCurrent = 0;
            private decimal _dPowerFactor = 0;
            private string _sHexString = "";

            public LogLine() { }
            public LogLine(DateTime LogTime, decimal Voltage, decimal Current, decimal PowerFactor, string HexString, DateTime FileTime)
            {
                _dtLogTime = LogTime;
                _dVoltage = Voltage;
                _dCurrent = Current;
                _dPowerFactor = PowerFactor;
                _sHexString = HexString;
                _dtFileTime = FileTime;
            }

            public DateTime FileTime { get { return _dtFileTime; } }
            public DateTime LogTime { get { return _dtLogTime; } }
            public string HexDataString { get { return _sHexString; } }
            public decimal Voltage { get { return _dVoltage; } }
            public decimal Current { get { return _dCurrent; } }
            public decimal PowerFactor { get { return _dPowerFactor; } }
            public decimal VA { get { return Voltage * Current; } }
            public decimal W { get { return VA * PowerFactor; } }

        }
 
        public List<LogLine> Log;
        private void ReadDataFile(string sFilename)
        {
            if (!File.Exists(sFilename)) return;
            BinaryReader binReader = new BinaryReader(File.Open(sFilename, FileMode.Open));
            
            //Check header (already done at ReadFile)
            byte[] bHeader = new byte[3];
            int iHCnt = binReader.Read(bHeader, 0, 3);

            //StartDatum logboek
            //3-5: MM/DD/YY
            //6-7: HH:MM
            int iMinutes = 0;
            DateTime dtStartDate = DateTime.Now;
            byte[] bDate = new byte[5];
            int iDCnt = binReader.Read(bDate, 0, 5);
            if (iDCnt == 5)
            {
                dtStartDate = new DateTime((int)bDate[2] + 2000, (int)bDate[0], (int)bDate[1], (int)bDate[3], (int)bDate[4], 0);
                System.Diagnostics.Debug.WriteLine("ElReader: DATA: " + dtStartDate.ToString());
            }
            else
            {
                System.Diagnostics.Debug.WriteLine("ElReader: DATA: Error reading startdate!");
                return;
            }
            DateTime dtFileStartDateTime = dtStartDate;
            
            //Go into loop, retreive and proces every time 5 bytes
            //VV:CC:[PF] (2bytes,2bytes,1byte)
            byte[] buffer = new byte[5];
            int iCnt = 0;
            while ((iCnt = binReader.Read(buffer, 0, 5)) == 5)
            {
                //Seperate bytes
                byte[] bVoltage = new byte[] { 0x00, 0x00, buffer[0], buffer[1] };
                byte[] bCurrent = new byte[] { 0x00, 0x00, buffer[2], buffer[3] };
                byte[] bPowerFactor = new byte[] { 0x00, 0x00, 0x00, buffer[4] };
                
                string hexDataString = "";
                
                //Google this one..... :) 
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(bVoltage);
                    Array.Reverse(bCurrent);
                    Array.Reverse(bPowerFactor);
                }

                //Translate bytes to decimals
                decimal Voltage = (decimal)((decimal)BitConverter.ToInt32(bVoltage, 0) / 10);
                decimal Current = (decimal)((decimal)BitConverter.ToInt32(bCurrent, 0) / 1000);
                decimal PowerFactor = (decimal)((decimal)BitConverter.ToInt32(bPowerFactor, 0) / 100);
                
                //Some additional calculations, not needed
                decimal VA = Voltage * Current; //NL: Schijnbaar
                decimal W = VA * PowerFactor;   //NL: Werkelijk opgenomen

                //HexString representing bytes,, for fun and debuging
                string HexString = string.Empty;
                foreach (byte b in buffer)
                {
                    HexString += string.Format("{0:X2}", b);
                }
                hexDataString = HexString;
                //For debuging, 
                System.Diagnostics.Debug.WriteLine("{0}: {6}: {1}V {2}A {3}PF {4}VA {5}W", HexString, Voltage, Current, PowerFactor, VA, W, dtStartDate.AddMinutes(iMinutes).ToShortTimeString());

                //Acceptable value ranges
                bool bVoltageOk = ((Voltage > 200) && (Voltage < 260));
                bool bCurrentOk = ((Current > 0) && (Current < 16));
                bool bPowerFactorOk = ((PowerFactor > (decimal)0.1) && (PowerFactor <= 1));
                bool bVAOk = ((VA > 0) && (VA < 3680));
                bool bWOk = ((W > 0) && (W < 3680));

                //Count value errors
                int iECnt = 0;
                if (!bVoltageOk)
                {
                    System.Diagnostics.Debug.WriteLine("\tElReader.Data.Voltage.OutOfRange: {0}", Voltage);
                    iECnt++;
                }
                if (!bCurrentOk)
                {
                    System.Diagnostics.Debug.WriteLine("\tElReader.Data.Current.OutOfRange: {0}", Current);
                    iECnt++;
                }
                if (!bPowerFactorOk)
                {
                    System.Diagnostics.Debug.WriteLine("\tElReader.Data.PowerFactor.OutOfRange: {0}", PowerFactor);
                    iECnt++;
                }

                //If any error in values....
                if (iECnt > 0 &&
                    (buffer[0] == 0xE0 &&       //It might be a new log
                    buffer[1] == 0xC5 &&        //This bytes say logtime is being given
                    buffer[2] == 0xEA))         //in the next 5 bytes. See datasheet conrad.
                {                    
                    Console.WriteLine("\tElReader.Data.NewLogTimeFound!");
                    //Loop reads 5 bytes, new start byte is only 3 bytes
                    //so move 2 bytes back in time
                    binReader.BaseStream.Position = binReader.BaseStream.Position - 2;
                    byte[] bCalTime = new byte[5];
                    int iX = binReader.Read(bCalTime, 0, 5);

                    if (iX == 5)
                    {
                        DateTime dtCalibrate = new DateTime((int)bCalTime[2] + 2000, (int)bCalTime[0], (int)bCalTime[1], (int)bCalTime[3], (int)bCalTime[4], 0);
                        System.Diagnostics.Debug.WriteLine("ElReader.Data.CalibrateTime: Old " + dtStartDate.AddMinutes(iMinutes).ToString());
                        System.Diagnostics.Debug.WriteLine("ElReader.Data.CalibrateTime: New " + dtCalibrate.ToString());
                        dtStartDate = dtCalibrate;
                        iMinutes = 0;
                    }
                    else
                    {
                        System.Diagnostics.Debug.WriteLine("ElReader.Data.CalibrateTime: Error reading time");
                    }
                }
                //On any other error...
                if (iECnt > 0)
                {
                    System.Diagnostics.Debug.WriteLine("\tHet gaat fout! {3}: {0}, {1}, {2}", Voltage, Current, PowerFactor, HexString);

                    //Einde bestand?? (Possible end of file)
                    if (buffer[0] == 0xFF && buffer[1] == 0xFF && buffer[2] == 0xFF)
                    {
                        System.Diagnostics.Debug.WriteLine("Einde bestand???");
                        break;
                    }
                }

                ////No error?? register line
                //if (iECnt == 0)
                //{
                    LogLine ll = new LogLine(dtStartDate.AddMinutes(iMinutes), Voltage, Current, PowerFactor,hexDataString,dtFileStartDateTime);
                    Log.Add(ll);
                //}


                //Every 5 bytes is a new minute.
                iMinutes++;
            }
            binReader.Close();
        }

    
        

    }
}
